/***************************************************************************
 *  CISILIA PROJECT: main.c
 *
 *  Tue Jan  7 23:11:39 2003
 *  Copyright  2002-2003  Cisilia Team <cisilia@cisiar.org>
 *
 ****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libintl.h>
#define _(String) gettext (String)
#define gettext_noop(String) String
#define N_(String) gettext_noop (String)

#include "../include/cisilia.h"

/*--------------------------------------------------*/

uchar *alfabeto[LONG_PASSWD / 2];       /* !< Vector de Alfabetos */
datoUser *datosUsers = NULL;    /* !< Estructura del Proceso Padre */
LMhash *LMhashes = NULL;        /* !< Estructura de cada Hijo */
indice *indice_LMhashes = NULL; /* !< Indice de Busqueda de LMHashes */
indice **matriz_LMhashes = NULL;        /* !< Matriz de Busqueda de LMHashes */
MDstruct MDContext;             /* !< MD4 Context Structure */
param parametros;               /* !< Parametros de linea de comando */
int LONG_ALFABETO = 0;          /* !< Cantidad de caracteres en el alfabeto */
datos_hijo datosHijo;           /* !< Datos estadisticos del proceso hijo */
estadisticas datosPadre;        /* !< Datos estadisticos del proceso padre */
int socktcp = 0;                /* !< Socket de escucha tcp en el padre y cliente en el hijo */
int sockudp = 0;                /* !< Socket de escucha udp en el padre y cliente en el hijo */
int sock_frontend = 0;          /* !< Socket del cliente UDP */
datos_hijo *vhijos;             /* !< Vector de datos de los hijos en el padre */
#if CONEXION_PERMANENTE == 0
int cliente = 0;                /* !< Socket del padre conectado con un hijo */
int clientes = 1;               /* !< Cantidad de sockets del padre conectados con los hijos */
#else
int *vclientes = NULL;          /* !< Sockets del padre conectados con cada hijo */
int clientes = 0;               /* !< Cantidad de  sockets del padre conectados con los hijos */
#endif
int pid;                        /* !< PID del proceso actual */
int jerarquia;                  /* !< PADRE o HIJO */
extern int errno;

/*----------------------------------------------------*/
// ! Funcion Principal
int main(int argc, char **argv)
{
    int rango_desde[LONG_PASSWD / 2] = { 0 };
    int rango_hasta[LONG_PASSWD / 2] = { 0 };   /* IMPORTANTE ASIGNAR VALOR INICIAL */
    int subpid;

/* INICIAR SUBSISTEMA DE LOCALIZACION */
#ifdef ENABLE_NLS
    setlocale(LC_ALL, "");
    bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
    textdomain(PACKAGE);
#endif

    pid = getpid();
    datosPadre.pid = pid;

    leer_opciones(argc, argv);
    inicioPrincipal();

    crackear_passwords_dic(parametros.arch_dict);

/* IMPORTANTE QUE EL PRIMER id_hijo DE FUERZA BRUTA SEA IGUAL A 1 */
    for (datosHijo.id_hijo = 1; datosHijo.id_hijo <= parametros.procesos;
         datosHijo.id_hijo++) {
            
        switch (subpid = fork()) {
        case -1:               /* ERROR fork() */
            error(ERROR_FORK);
            break;
        
        case 0:                /* SUBPROCESO */
            inicioSubproc();
            CRACKEAR(rango_desde, rango_hasta);
            enviar_datos_hijo();
//          imprimir_estadisticas();
            salir();
            
        default:               /* PROCESO PADRE */
            mensaje("(PID:%d) Subproceso #%d (PID:%d)\n", datosPadre.pid,
                    datosHijo.id_hijo, subpid);
            escuchar();
        }
    }

    inicioPadre();
    cicloPrincipal();
    return 0;
}

/*---------------------------------------------------*/

void cicloPrincipal()
{
    do {
        escuchar();
    }
    while (clientes > 0 && datosPadre.cracks < datosPadre.lineas);

    /* DESACTIVAR LA ALARMA */
    signal(SIGALRM, SIG_IGN);
    ualarm(0, 0);

    enviar_estadisticas();
    mensaje("\n==>  F I N  <==\n\a");
    salir();
}

/*---------------------------------------------------*/
// ! Levantar a memoria el archivo de claves
int archivo_a_memoria(uchar * archivo)
{
    FILE *pfile;
    uchar linea[250];
    uchar *ptok;
    int i = 0, l = 0;
    uchar hash[16];

    if ((pfile = fopen(archivo, "rt")) == NULL) {
        error(ERROR_ARCH_PASSWD);
    }

    while (fgets(linea, 250, pfile) != NULL) {
        l++;
        if (strlen(linea) != 0) /* se puede preguntar por pstrtok==NULL?? */
            if (linea[0] != '#') {
                datosUsers =
                    (datoUser *) realloc(datosUsers,
                                         sizeof(datoUser) * (i + 1));
                if (datosUsers == NULL)
                    error(ERROR_REALLOC);

                LMhashes =
                    (LMhash *) realloc(LMhashes,
                                       sizeof(LMhash) * (2 * i + 2));
                if (LMhashes == NULL)
                    error(ERROR_REALLOC);

                ptok = strtok(linea, ":");
                if (ptok == NULL) {
                    mensaje("%s:%d: ERROR: Error en formato de la linea\n",
                            archivo, l);
                    continue;
                    error(ERROR_ARCH_PASSWD);
                }
                // mensaje("1er token = %s\n",ptok);
                memcpy(datosUsers[i].nombre, ptok, 20);

                datosUsers[i].
                    nombre[sizeof(datosUsers[i].nombre) - 1] = '\0';

                ptok = strtok(NULL, ":");
                if (ptok == NULL) {
                    mensaje("%s:%d: ERROR: Error en formato de la linea\n",
                            archivo, l);
                    continue;
                    error(ERROR_ARCH_PASSWD);
                }
                // mensaje("2do token = %s\n",ptok);

                ptok = strtok(NULL, ":");
                if (ptok == NULL) {
                    mensaje("%s:%d: ERROR: Error en formato de la linea\n",
                            archivo, l);
                    continue;
                    error(ERROR_ARCH_PASSWD);
                }
                // mensaje("3er token = %s\n",ptok);
                sscanf(ptok,
                       "%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X",
                       (uint *) & LMhashes[2 * i].LMsemihash[0],
                       (uint *) & LMhashes[2 * i].LMsemihash[1],
                       (uint *) & LMhashes[2 * i].LMsemihash[2],
                       (uint *) & LMhashes[2 * i].LMsemihash[3],
                       (uint *) & LMhashes[2 * i].LMsemihash[4],
                       (uint *) & LMhashes[2 * i].LMsemihash[5],
                       (uint *) & LMhashes[2 * i].LMsemihash[6],
                       (uint *) & LMhashes[2 * i].LMsemihash[7],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[0],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[1],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[2],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[3],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[4],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[5],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[6],
                       (uint *) & LMhashes[2 * i + 1].LMsemihash[7]);

                ptok = strtok(NULL, ":");
                if (ptok == NULL) {
                    mensaje("%s:%d: ERROR: Error en formato de la linea\n",
                            archivo, l);
                    continue;
                    error(ERROR_ARCH_PASSWD);
                }
                // mensaje("4to token = %s\n",ptok);
                sscanf(ptok,
                       "%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X%2X",
                       (uint *) & hash[0], (uint *) & hash[1],
                       (uint *) & hash[2], (uint *) & hash[3],
                       (uint *) & hash[4], (uint *) & hash[5],
                       (uint *) & hash[6], (uint *) & hash[7],
                       (uint *) & hash[8], (uint *) & hash[9],
                       (uint *) & hash[10], (uint *) & hash[11],
                       (uint *) & hash[12], (uint *) & hash[13],
                       (uint *) & hash[14], (uint *) & hash[15]);
                memcpy(datosUsers[i].NThash, hash, 16);

                LMhashes[2 * i].usuario = i;
                LMhashes[2 * i + 1].usuario = i;
                LMhashes[2 * i].mitad = IZQ;
                LMhashes[2 * i + 1].mitad = DER;
                datosUsers[i].ok = 0;
                memset(datosUsers[i].passwd, '\0', LONG_PASSWD + 1);
                i++;
            }
    }
    fclose(pfile);

    if (i == 0)
        error(ERROR_ARCH_PASSWD);

    mensaje("Usuarios leidos: %d\n", i);
    return i;
}

/*-----------------------------------------*/
// ! Lee el archivo del Alfabeto o carga uno por defecto
/*! \param nombre_alfabeto Nombre del archivo 
    \return Codigo de resultado 
*/
int cargar_alfabeto(uchar * nombre_alfabeto)
{
    FILE *arch_alfabeto;
    int i, j, k, c;
/* ARREGLOS AUXILIARES PARA CARGAR ALFABETO DE ARCHIVO */
    char rep[LONG_ALF_ASCII] = { 0 };
    uchar alf[LONG_ALF_ASCII];

/* CARGAR ALFABETO DE ARCHIVO */
    if (nombre_alfabeto != NULL) {
        if ((arch_alfabeto = fopen(nombre_alfabeto, "rt")) == NULL)
            error(ERROR_ARCH_ALFABETO);
        do {
            c = toupper(fgetc(arch_alfabeto));

            if (isprint(c)) {
                if (rep[c] == 0) {
                    rep[c] = 1;
                    alf[LONG_ALFABETO] = c;
                    LONG_ALFABETO++;
                }
            }
        }
        while (c != EOF);

        fclose(arch_alfabeto);

#if POS_BARRA_0 >= 0
        LONG_ALFABETO++;        /* PARA CONTAR EL '\0' */
#endif

        for (i = 0; i < LONG_PASSWD / 2; i++)
            if ((alfabeto[i] =
                 (uchar *) malloc(LONG_ALFABETO * sizeof(uchar))) == NULL)
                error(ERROR_MALLOC);

        for (i = 0; i < LONG_PASSWD / 2; i++) {
#if POS_BARRA_0 >= 0
            alfabeto[i][POS_BARRA_0] = BARRA_0;
#endif
            for (j = 0, k = 0; j < LONG_ALFABETO; j++, k++)
                if (j != POS_BARRA_0) {
                    alfabeto[i][j] = alf[k];
                    // mensaje("alfabeto[%d][%d]=%c
                    // \talf[%d]=%c\n",i,j,alfabeto[i][j],k,alf[k]);
                } else
                    k--;
        }
    } else {                    /* USAR ALFABETO POR DEFECTO */
        if (POS_BARRA_0 >= 0)
            LONG_ALFABETO = 27;
        else
            LONG_ALFABETO = 26;

#if ALFABETO_NUMERICO != 0
        LONG_ALFABETO += 10;
#endif

        for (i = 0; i < LONG_PASSWD / 2; i++)
            if ((alfabeto[i] =
                 (uchar *) malloc(LONG_ALFABETO * sizeof(uchar))) == NULL)
                error(ERROR_MALLOC);

        for (i = 0; i < LONG_PASSWD / 2; i++) {
            if (POS_BARRA_0 >= 0)
                alfabeto[i][POS_BARRA_0] = BARRA_0;

            j = 0;
            for (k = 'A'; k <= 'Z'; k++, j++)
                if (j != POS_BARRA_0)
                    alfabeto[i][j] = toupper(k);
                else
                    k--;

#if ALFABETO_NUMERICO != 0      /* AGREGAR CARACTERES NUMERICOS AL ALFABETO */
            for (k = '0'; k <= '9'; k++, j++)
                if (j != POS_BARRA_0)
                    alfabeto[i][j] = k;
                else
                    k--;
#endif
        }
    }
    mensaje("Longitud alfabeto: %d\n\n", LONG_ALFABETO);
    for (i = 0; i < LONG_PASSWD / 2; i++) {
        for (j = 0; j < LONG_ALFABETO; j++)
            mensaje("%c", alfabeto[i][j]);
        mensaje("\n");
    }
    mensaje("\n");
    return OK;
}

/*-----------------------------------------*/
// ! Imprime la los datos de la estructura de usuarios
/*
int imprimir_datos_users_plano()
{
    int w, j;

    mensaje("\n");
    for (w = 0; w < datosPadre.lineas; w++) {
	mensaje("%s:    \t", datosUsers[w].nombre);
	for (j = 0; j < 8; j++)
	    mensaje("%02X", LMhashes[2 * w].LMsemihash[j]);
	mensaje(" (%02d:%02d:%02d) ", LMhashes[2 * w].usuario,
		LMhashes[2 * w].mitad, 2 * w);
	for (j = 0; j < 8; j++)
	    mensaje("%02X", LMhashes[2 * w + 1].LMsemihash[j]);
	mensaje(" (%02d:%02d:%02d) ", LMhashes[2 * w + 1].usuario,
		LMhashes[2 * w + 1].mitad, 2 * w + 1);
	// for(j=0;j<16;j++)
	// mensaje("%02X",datosUsers[w].NThash[j]);
	mensaje("\n");
    }
    return OK;
}
*/
/*---------------------------------------*/
// ! Ordena los semiLMhashes usando el algoritmo quicksort
int ordenar_LMhashes(int largo)
{
    qsort(LMhashes, largo, sizeof(LMhash), &comparar_LMhashes);
    return OK;
}

/*----------------------------------------*/
// ! Funcion que utiliza quicksort para ordenar los semiLMhashes
int comparar_LMhashes(const void *x, const void *y)
{
    return (memcmp
            (((LMhash *) x)->LMsemihash, ((LMhash *) y)->LMsemihash, 8));
}

/*-----------------------------------------*/
// ! Crea el indice de busqueda de semiLMhashes
int generar_indice_LMhashes(int largo)
{
    int i, j;
    uchar LMhash_0_ant;
    uchar LMhash_1_ant;

/* GENERAR INDICE DE BUSQUEDA DE 1 DIMENSION */

    indice_LMhashes = (indice *) malloc(sizeof(indice) * LONG_ALF_ASCII);
    if (indice_LMhashes == NULL)
        error(ERROR_MALLOC);

    for (i = 0; i < LONG_ALF_ASCII; i++) {
        indice_LMhashes[i].pos = HASH_INEXISTENTE;
        indice_LMhashes[i].cant = 0;
    }

    LMhash_0_ant = LMhashes[0].LMsemihash[0];
    indice_LMhashes[LMhash_0_ant].pos = 0;
    indice_LMhashes[LMhash_0_ant].cant++;

    for (i = 1; i < largo; i++) {
        if (LMhashes[i].LMsemihash[0] != LMhash_0_ant) {
            LMhash_0_ant = LMhashes[i].LMsemihash[0];
            indice_LMhashes[LMhash_0_ant].pos = i;
        }
        indice_LMhashes[LMhash_0_ant].cant++;
    }

/* GENERAR INDICE DE BUSQUEDA DE 2 DIMENSIONES */

    matriz_LMhashes =
        (indice **) malloc(sizeof(indice *) * LONG_ALF_ASCII);
    if (matriz_LMhashes == NULL)
        error(ERROR_MALLOC);

    for (i = 0; i < LONG_ALF_ASCII; i++) {
        matriz_LMhashes[i] =
            (indice *) malloc(sizeof(indice) * LONG_ALF_ASCII);
        if (matriz_LMhashes[i] == NULL)
            error(ERROR_MALLOC);
    }

    for (i = 0; i < LONG_ALF_ASCII; i++) {
        for (j = 0; j < LONG_ALF_ASCII; j++) {
            matriz_LMhashes[i][j].pos = HASH_INEXISTENTE;
            matriz_LMhashes[i][j].cant = 0;
        }
    }

    LMhash_0_ant = LMhashes[0].LMsemihash[0];
    LMhash_1_ant = LMhashes[0].LMsemihash[1];
    matriz_LMhashes[LMhash_0_ant][LMhash_1_ant].pos = 0;
    matriz_LMhashes[LMhash_0_ant][LMhash_1_ant].cant++;

    for (i = 1; i < largo; i++) {
        if (LMhashes[i].LMsemihash[0] != LMhash_0_ant ||
            LMhashes[i].LMsemihash[1] != LMhash_1_ant) {
            LMhash_0_ant = LMhashes[i].LMsemihash[0];
            LMhash_1_ant = LMhashes[i].LMsemihash[1];
            matriz_LMhashes[LMhash_0_ant][LMhash_1_ant].pos = i;
        }
        matriz_LMhashes[LMhash_0_ant][LMhash_1_ant].cant++;
    }
//    for(j=0;j<LONG_ALF_ASCII;j++)
//      mensaje("indice[%d=0x%02X] = %d;%d\n",j,j,indice_LMhashes[j].pos,indice_LMhashes[j].cant);
//    for(i=0;i<LONG_ALF_ASCII;i++)
//      for(j=0;j<LONG_ALF_ASCII;j++)
//          mensaje("matriz[%d=0x%02X][%d=0x%02X] = %d;%d\n",i,i,j,j,matriz_LMhashes[i][j].pos,matriz_LMhashes[i][j].cant);
    return OK;
}

/*-----------------------------------------*/
// ! Calcula la cantidad total de claves validas a generar
double total_palabras_validas(void)
{
    int i;
    double r = 0;
    for (i = 0; i <= LONG_PASSWD / 2; i++)
        r += pow(LONG_ALFABETO - 1, i);
    return r;
}

/*-----------------------------------------*/
// ! Calcula la cantidad total de claves a generar
/*
double calcular_rango_total(void)
{
    return (pow(LONG_ALFABETO, LONG_PASSWD / 2) - 1);
}
*/
/*-----------------------------------------*/
// ! Calcula cuantas claves procesara cada proceso hijo
/*
double dividir_rango_total(int cant, int intervalo[LONG_PASSWD / 2])
{
    double rango_total = calcular_rango_total();
    double rango = rango_total / cant;
    int i;
    mensaje("RANGO TOTAL = %.1f\n", rango_total);
    mensaje("INTERVALO   = %.1f\n", rango);

    cambiar_base(rango, LONG_ALFABETO, intervalo, LONG_PASSWD / 2);

    mensaje("INTERVALO   = ");
    for (i = 0; i < LONG_PASSWD / 2; i++)
	mensaje("%3d ", intervalo[i]);
*/
/* VERIFICACION */
/*    rango = 0;
    for (i = 0; i < LONG_PASSWD / 2; i++)
	rango +=
	    intervalo[i] * pow(LONG_ALFABETO, LONG_PASSWD / 2 - i - 1);
    mensaje("= %.1f\n\n", rango);
    return floor(rango);
}
*/
/*------------------------------------------*/
// ! Cambia un numero en base 10 a cualquier base
/*
int cambiar_base(double nro, int base, int *res, int largo)
{
    int i;
    for (i = largo - 1; i >= 0; i--) {
	res[i] = (unsigned long long) nro % base;
	nro = floor(nro / base);
    }
    return OK;
}
*/
/*------------------------------------------*/
// ! Calcula el siguiente rango a procesar sumando el intervalo al rango anterior
/*
int
asignar_rango(int desde[LONG_PASSWD / 2], int hasta[LONG_PASSWD / 2],
	      const int intervalo[LONG_PASSWD / 2])
{
    int i;
    for (i = 0; i < LONG_PASSWD / 2; i++)
	desde[i] = hasta[i];
    sumar_base_n(desde, intervalo, LONG_ALFABETO, LONG_PASSWD / 2, hasta);
#if SUBPROC_GEN_PAL_NULA != 0
    sumar_uno_base_n(desde, LONG_ALFABETO, LONG_PASSWD / 2);
#endif
    mensaje("\nRANGO %d: ",datosHijo.id_hijo);
    for (i = 0; i < LONG_PASSWD / 2; i++)
	mensaje("%2d ", desde[i]);
    mensaje(" ==> ");
    for (i = 0; i < LONG_PASSWD / 2; i++)
	mensaje("%2d ", hasta[i]);
    return OK;
}
*/
/*-------------------------------------------*/
// ! Suma dos numeros en una base numerica determinada
/*
int
sumar_base_n(const int n1[], const int n2[], const int base,
	     const int largo, int res[])
{
    int i, k, resto = 0;

    for (i = largo - 1; i >= 0; i--) {
	k = n1[i] + n2[i] + resto;
	if (k < base) {
	    res[i] = k;
	    resto = 0;
	} else {
	    res[i] = k - base;
	    resto = 1;
	}
    }
    return OK;
}
*/
/*--------------------------------------------*/
// ! Incrementa en uno un numero en una determinada base
/*
static inline int
sumar_uno_base_n(int n1[], const int base, const int largo)
{
    int i;

    for (i = largo - 1; i >= 0; i--)
	if (++n1[i] < base)
	    return OK;
	else
	    n1[i] = 0;

    return OK;
}
*/
/*---------------------------------------------*/
// ! Inicializa todas las posiciones de un rango a un mismo valor
/*
int init_rango(int *vector, int valor)
{
    int i;

    for (i = 0; i < LONG_PASSWD / 2; i++)
	vector[i] = valor;

    return OK;
}
*/
/*-------------------------------------------*/
// ! Algoritmo de busqueda binaria de un semiLMhash en el vector de semiLMhashes
int
buscar_LMhash_bin(const uchar * semiLMhash, const uchar * semipasswd,
                  int inicio, int final)
{
    int pos;
    int cmp;
    int user;
    char mitad;
    int pto_medio;

    pos = (inicio + final) / 2;

    while (((cmp = memcmp(LMhashes[pos].LMsemihash, semiLMhash, 8)) != 0)
           && (inicio <= final)) {
        if (cmp < 0)
            inicio = pos + 1;
        else
            final = pos - 1;

        pos = (inicio + final) / 2;
    }

    if (cmp == 0) {
        pto_medio = pos;
        while (pos <= final
               && memcmp(LMhashes[pos].LMsemihash, semiLMhash, 8) == 0) {
            user = LMhashes[pos].usuario;
            mitad = LMhashes[pos].mitad;
            enviarDatosUsers(semipasswd, user, mitad);
//printf ("MITAD ENCONTRADA en %d user=%d mitad=%d\n", pos, user, mitad);
            pos++;
        }

        pos = pto_medio - 1;
        while (pos >= inicio) {
            if (memcmp(LMhashes[pos].LMsemihash, semiLMhash, 8) == 0) {
                user = LMhashes[pos].usuario;
                mitad = LMhashes[pos].mitad;
                enviarDatosUsers(semipasswd, user, mitad);
//printf ("MITAD ENCONTRADA en %d user=%d mitad=%d\n", pos, user, mitad);
            } else
                break;
            pos--;
        }
        return (ENCONTRADO);
    } else {
        // mensaje("MITAD NO ENCONTRADA\n");
        return (!ENCONTRADO);
    }
}

/*-------------------------------------------*/
// ! Algoritmo de busqueda binaria de un semiLMhash en el vector de semiLMhashes
/*
int
buscar_LMhash_bin_total(const uchar * semiLMhash, const uchar * semipasswd)
{
    buscar_LMhash_bin(semiLMhash, semipasswd, 0, 2 * datosPadre.lineas - 1);
    return OK;
}
*/
/*-------------------------------------------*/
// ! Algoritmo de busqueda indexada de un semiLMhash en el vector de semiLMhashes
/*
int
buscar_LMhash_indice(const uchar * semiLMhash, const uchar * semipasswd)
{
    int user;
    char mitad;
    int encontrada = 0;
    int pos = indice_LMhashes[semiLMhash[0]].pos;

    for (;
	 pos <= datosPadre.lineas * 2 - 1
	 && LMhashes[pos].LMsemihash[0] == semiLMhash[0]; pos++) {
	if (memcmp(LMhashes[pos].LMsemihash, semiLMhash, 8) == 0) {
	    user = LMhashes[pos].usuario;
	    mitad = LMhashes[pos].mitad;
	    enviarDatosUsers(semipasswd, user, mitad);
	    // mensaje("MITAD ENCONTRADA en %d
	    // user=%d\n",pos,user);
	    encontrada = 1;
	} else {
	    if (encontrada)
		return ENCONTRADO;
	}
    }
    return !ENCONTRADO;
}
*/
/*---------------------------------------------*/
// ! Algoritmo de busqueda hibrida
/*
int
buscar_LMhash_indice_bin(const uchar * semiLMhash,
			 const uchar * semipasswd)
{
    int pos = indice_LMhashes[semiLMhash[0]].pos;
    int final;

    if (pos == HASH_INEXISTENTE)
	return !ENCONTRADO;

    for (final = semiLMhash[0] + 1;
	 indice_LMhashes[final].pos == HASH_INEXISTENTE
	 && final < LONG_ALF_ASCII; final++);

    if (final == LONG_ALF_ASCII)
	return (buscar_LMhash_bin
		(semiLMhash, semipasswd, pos, datosPadre.lineas * 2 - 1));
    else
	return (buscar_LMhash_bin
		(semiLMhash, semipasswd, pos, indice_LMhashes[final].pos));
}
*/
/*--------------------------------------------*/
// ! Algoritmo de busqueda hibrida optimizado
int
buscar_LMhash_indice_bin_extra(const uchar * semiLMhash,
                               const uchar * semipasswd)
{
    int pos = indice_LMhashes[semiLMhash[0]].pos;
    int cant = indice_LMhashes[semiLMhash[0]].cant;

    if (pos == HASH_INEXISTENTE)
        return !ENCONTRADO;

    return (buscar_LMhash_bin
            (semiLMhash, semipasswd, pos, pos + cant /* - 1 */ ));
}

/*--------------------------------------------*/
// ! Algoritmo de busqueda hibrida en la matriz de dos entradas
int
buscar_LMhash_matriz_bin(const uchar * semiLMhash,
                         const uchar * semipasswd)
{
    int pos = matriz_LMhashes[semiLMhash[0]][semiLMhash[1]].pos;
    int cant = matriz_LMhashes[semiLMhash[0]][semiLMhash[1]].cant;

    if (pos == HASH_INEXISTENTE)
        return !ENCONTRADO;

    return (buscar_LMhash_bin
            (semiLMhash, semipasswd, pos, pos + cant /* - 1 */ ));
}

/*--------------------------------------------*/

int
enviarDatosUsers(const uchar * semipasswd, const int user, const int mitad)
{
    int tx = 0;
    int tx_total = 0;
    crack descifrado;

/* VERIFICAR INTEGRIDAD DE LOS DATOS */
    if ((user < 0) || (user > datosPadre.lineas - 1)) {
        advertencia("HIJO: Usuario invalido");
//      _exit(1);
        return !OK;
    }
    if ((mitad != IZQ) && (mitad != DER)) {
        advertencia("HIJO: Mitad invalida");
//      _exit(1);
        return !OK;
    }

    descifrado.usuario = user;
    descifrado.mitad = mitad;
    descifrado.pid = datosHijo.pid;
    memcpy(descifrado.semipasswd, semipasswd,
           sizeof(uchar) * LONG_PASSWD / 2);
#if CONEXION_PERMANENTE == 0
    socktcp = conectarTCP(LOCALHOST, parametros.puerto);
#endif

    do {
        errno = 0;
        tx = send(socktcp, &(((char *) &descifrado)[tx_total]),
                  sizeof(descifrado) - tx_total, 0);
        tx_total += tx;
    } while (tx_total < sizeof(descifrado) && tx > 0);

#if CONEXION_PERMANENTE == 0
    close(socktcp);
#endif
    if (tx_total < sizeof(descifrado)) {
        mensaje("tx=%d de %d ", tx_total, sizeof(descifrado));
        advertencia("ERROR enviando datos de usuario");
        return !OK;
    }
    return OK;
}

/*-------------------------------------------*/

int
recibirDatosUsers(const uchar * semipasswd, const int user,
                  const char mitad)
{
/* VERIFICAR INTEGRIDAD DE LOS DATOS */
    if ((user < 0) || (user > datosPadre.lineas - 1)) {
//      fprintf(stderr,"user:%d\n",user);
//      imprimir_estado_padre ();
        advertencia("Usuario invalido");
//      _exit(1);
        return !OK;
    }
    if ((mitad != IZQ) && (mitad != DER)) {
//      fprintf(stderr,"mitad:%d\n",mitad);
//      imprimir_estado_padre ();
        advertencia("Mitad invalida");
//      _exit(1);
        return !OK;
    }
/* VERIFICAR QUE NO ES UNA SEMICLAVE REPETIDA */
//mensaje("Usuario valido0\n");
    if ((datosUsers[user].ok < PASSWD_OK)
        && (datosUsers[user].ok != mitad)) {
        datosPadre.semicracks++;
        datosUsers[user].ok += mitad;
        memcpy(&
               (datosUsers[user].
                passwd[mitad == IZQ ? 0 : LONG_PASSWD / 2]), semipasswd,
               LONG_PASSWD / 2);
        return OK;
    } else
        return !OK;
}

/*-------------------------------------------*/

int chequear_passwd_ok(const int user)
{
    int largo;
    double s;
    unsigned long h, m;
    int nt;
    int tx;
    datoCrack cracked;

    gettimeofday(&(datosPadre.t), NULL);
    s = (datosPadre.t.tv_sec - datosPadre.ti.tv_sec) +
        (datosPadre.t.tv_usec - datosPadre.ti.tv_usec) * .000001;
    m = s / 60;
    h = m / 60;

    if (datosUsers[user].ok == PASSWD_OK) {
        datosPadre.cracks++;
        memcpy(cracked.LMpasswd, datosUsers[user].passwd, LONG_PASSWD + 1);
        largo = strlen(datosUsers[user].passwd);
        nt = calcularPwdMD_2(datosUsers[user].passwd, 0, largo, user);

        cracked.cracks = datosPadre.cracks;
        cracked.ti = datosPadre.ti;
        cracked.t = datosPadre.t;
        memcpy(cracked.nombre, datosUsers[user].nombre, 20);
        if (nt == ENCONTRADO)
            memcpy(cracked.NTpasswd, datosUsers[user].passwd,
                   LONG_PASSWD + 1);
        else
            memcpy(cracked.NTpasswd, "???", LONG_PASSWD + 1);

        if (datosPadre.cracks == 1)
            logear(_("#\tTIME\t\tUSER\t\tLMPASSWD\tNTPASSWD\n"));

        logear("%lld\t%02lu:%02lu:%06.3f\t%s\t%s\t%s\n",
               datosPadre.cracks, h, m - h * 60, s - m * 60,
               datosUsers[user].nombre, cracked.LMpasswd,
               cracked.NTpasswd);

        if (sock_frontend > 0) {
            tx = send(sock_frontend, &cracked, sizeof(cracked), 0);
//mensaje("TX ");
            if (tx < sizeof(cracked)) {
                advertencia("ERROR en send chequeando password");
                return !OK;
            }
        }
        return OK;
    } else
        return !OK;
}

/*-------------------------------------------*/
// ! Genera y crackea claves en un rango
/*
int
crackear_passwords_4(const int desde[LONG_PASSWD / 2],
		     const int hasta[LONG_PASSWD / 2])
{
    uchar semipasswd[LONG_PASSWD / 2 ] = { '\0' };
    uchar semiLMhash[8];
    int i[LONG_PASSWD / 2];
    int vpos[LONG_PASSWD / 2];
    int j;
    
    for (j = 0; j < LONG_PASSWD / 2; j++)
	vpos[j] = desde[j];

    for (i[6] = vpos[0]; i[6] < LONG_ALFABETO; i[6]++) {
	semipasswd[6] = alfabeto[6][i[6]];
	for (i[5] = vpos[1]; i[5] < LONG_ALFABETO; i[5]++) {
	    if (i[6] != POS_BARRA_0 && i[5] == POS_BARRA_0)
		continue;
	    semipasswd[5] = alfabeto[5][i[5]];
	    for (i[4] = vpos[2]; i[4] < LONG_ALFABETO; i[4]++) {
		if (i[5] != POS_BARRA_0 && i[4] == POS_BARRA_0)
		    continue;
		semipasswd[4] = alfabeto[4][i[4]];
		for (i[3] = vpos[3]; i[3] < LONG_ALFABETO; i[3]++) {
		    if (i[4] != POS_BARRA_0 && i[3] == POS_BARRA_0)
			continue;
		    semipasswd[3] = alfabeto[3][i[3]];
		    for (i[2] = vpos[4]; i[2] < LONG_ALFABETO; i[2]++) {
			if (i[3] != POS_BARRA_0 && i[2] == POS_BARRA_0)
			    continue;
			semipasswd[2] = alfabeto[2][i[2]];
			for (i[1] = vpos[5]; i[1] < LONG_ALFABETO; i[1]++) {
			    if (i[2] != POS_BARRA_0 && i[1] == POS_BARRA_0)
				continue;
			    semipasswd[1] = alfabeto[1][i[1]];
			    for (i[0] = vpos[6]; i[0] < LONG_ALFABETO;
				 i[0]++) {
				if (i[1] != POS_BARRA_0 && i[0] == POS_BARRA_0)
				    continue;
				if (i[6] > hasta[0])
				    return OK;
				if (i[6] == hasta[0]) {
				    if (i[5] > hasta[1])
					return OK;
				    if (i[5] == hasta[1]) {
					if (i[4] > hasta[2])
					    return OK;
					if (i[4] == hasta[2]) {
					    if (i[3] > hasta[3])
						return OK;
					    if (i[3] == hasta[3]) {
						if (i[2] > hasta[4])
						    return OK;
						if (i[2] == hasta[4]) {
						    if (i[1] > hasta[5])
							return OK;
						    if (i[1] == hasta[5]) {
							if (i[0] >
							    hasta[6])
							    return OK;
						    }
						}
					    }
					}
				    }
				}

				semipasswd[0] = alfabeto[0][i[0]];
//                              printf("=> %010llu- (datosHijo.pid:%d) %s\n",datosHijo.palabras, datosHijo.pid, semipasswd);
//                              fflush(stdout);
//                               usleep(1000);

//				   for (j=0;j<LONG_PASSWD / 2;j++){
//				   if(i[j] > hasta[j]) return OK; } 

				E1(semipasswd, CONSTANTE_MAGICA,
				   semiLMhash);
				BUSCAR(semiLMhash, semipasswd);
				datosHijo.palabras++;
			    }
			    vpos[6] = 0;
			}
			vpos[5] = 0;
		    }
		    vpos[4] = 0;
		}
		vpos[3] = 0;
	    }
	    vpos[2] = 0;
	}
	vpos[1] = 0;
    }
    return OK;
}
*/
/*-------------------------------------------*/
// ! Genera y crackea claves en un rango
int
crackear_passwords_5(const int desde[LONG_PASSWD / 2],
                     const int hasta[LONG_PASSWD / 2])
{
    uchar semipasswd[LONG_PASSWD / 2 /*+ 1 */ ] = { '\0' };
    uchar semiLMhash[8];
    int i[LONG_PASSWD / 2];
//    int vpos[LONG_PASSWD / 2];
    unsigned long long orden = 0;

//    cambiar_base(datosHijo.id_hijo, LONG_ALFABETO, vpos, LONG_PASSWD / 2);

    for (i[6] = 0; i[6] < LONG_ALFABETO; i[6]++) {
        semipasswd[6] = alfabeto[6][i[6]];
        for (i[5] = 0; i[5] < LONG_ALFABETO; i[5]++) {
            if (i[6] != POS_BARRA_0 && i[5] == POS_BARRA_0)
                continue;
            semipasswd[5] = alfabeto[5][i[5]];
            for (i[4] = 0; i[4] < LONG_ALFABETO; i[4]++) {
                if (i[5] != POS_BARRA_0 && i[4] == POS_BARRA_0)
                    continue;
                semipasswd[4] = alfabeto[4][i[4]];
                for (i[3] = 0; i[3] < LONG_ALFABETO; i[3]++) {
                    if (i[4] != POS_BARRA_0 && i[3] == POS_BARRA_0)
                        continue;
                    semipasswd[3] = alfabeto[3][i[3]];
                    for (i[2] = 0; i[2] < LONG_ALFABETO; i[2]++) {
                        if (i[3] != POS_BARRA_0 && i[2] == POS_BARRA_0)
                            continue;
                        semipasswd[2] = alfabeto[2][i[2]];
                        for (i[1] = 0; i[1] < LONG_ALFABETO; i[1]++) {
                            if (i[2] != POS_BARRA_0 && i[1] == POS_BARRA_0)
                                continue;
                            semipasswd[1] = alfabeto[1][i[1]];
                            for (i[0] = 0; i[0] < LONG_ALFABETO;
//                               i[0]+= parametros.procesos) {
                                 i[0]++) {
                                if (i[1] !=
                                    POS_BARRA_0 && i[0] == POS_BARRA_0)
                                    continue;

//                              printf("hijo=%d orden=%llu ordenmodproc=%llu hijo-1=%d\n",datosHijo.id_hijo,orden,orden%parametros.procesos,datosHijo.id_hijo-1);
//                              fflush(stdout);
                                if ((orden % parametros.procesos) ==
                                    (datosHijo.id_hijo - 1)) {
                                    semipasswd[0] = alfabeto[0][i[0]];
//                                  printf("%2d\t%llu\t%s\n",datosHijo.id_hijo,orden,semipasswd);
//                                  fflush(stdout);
//                                  usleep(1000);
                                    E1(semipasswd, CONSTANTE_MAGICA,
                                       semiLMhash);
                                    BUSCAR(semiLMhash, semipasswd);
                                    datosHijo.palabras++;
                                }
                                orden++;
                            }
//                          vpos[6] = i[0] - LONG_ALFABETO;
                        }
                    }
                }
            }
        }
    }
    return OK;
}

/*-------------------------------------------*/
// ! Imprime un codigo de error y aborta la ejecucion 
void error(int codigo)
{
    fprintf(stderr, "(PID:%d) ", pid);
    switch (codigo) {
    case ERROR_FORK:
        perror("ERROR en fork()");
        break;
    case ERROR_MALLOC:
        perror("ERROR en malloc()");
        break;
    case ERROR_ARCH_PASSWD:
        perror("ERROR en archivo de passwords");
        break;
    case ERROR_ARCH_ALFABETO:
        perror("ERROR en archivo de alfabeto");
        break;
    case ERROR_ARCH_DIC:
        perror("ERROR en archivo diccionario");
        break;
    case ERROR_REALLOC:
        perror("ERROR en realloc()");
        break;
    case ERROR_SOCKET:
        perror("ERROR en socket()");
        break;
    case ERROR_SOCKET_SEND:
        perror("ERROR en send()");
        break;
    case ERROR_SOCKET_RECV:
        perror("ERROR en recv()");
        break;
    case ERROR_SOCKET_BIND:
        perror("ERROR en bind()");
        break;
    case ERROR_SOCKET_LISTEN:
        perror("ERROR en listen()");
        break;
    case ERROR_SOCKET_HOSTNAME:
        perror("ERROR en gethostbyname()");
        break;
    case ERROR_SOCKET_ACCEPT:
        perror("ERROR en accept()");
        break;
    case ERROR_SOCKET_CONN:
        perror("ERROR en connect()");
        break;
    case ERROR_SOCKET_SELECT:
        perror("ERROR en select()");
        break;
    case ERROR_SOCKET_SIGPIPE:
        perror("ERROR en la conexion");
        break;
    default:
        perror("ERROR NO CONTEMPLADO");
    }
    mensaje("\a");
    salir();
}

/*-------------------------------------------*/
// ! Calcula las combinaciones entre MAY/min de una clave y genera el NThash
/*---------------------------------------------
* Muchachos, esta funcion lo que hace es armar un nuevo
* alfabeto, pero con la password encontrada en mayusculas
* y minusculas. Y luego calcula las variaciones de las
* mismas. Es por eso que LM la devuelve en MAYUSCULAS
* y luego usamos este rango acotado para calcular con MD4
* ---------------------------------------------*/
static inline int
calcularPwdMD_2(uchar * passwd, int pos, int fin, int user)
{
    uchar NThash[16];
    if (pos == fin) {
        calcularNThash(passwd, NThash);
        if (memcmp(datosUsers[user].NThash, NThash, 16) == 0)
            return ENCONTRADO;
        else
            return !ENCONTRADO;
    }

    passwd[pos] = tolower(passwd[pos]);
    if (calcularPwdMD_2(passwd, pos + 1, fin, user) == ENCONTRADO)
        return ENCONTRADO;
    passwd[pos] = toupper(passwd[pos]);
    if (calcularPwdMD_2(passwd, pos + 1, fin, user) == ENCONTRADO)
        return ENCONTRADO;
    return !ENCONTRADO;
}

/*-------------------------------*/
// ! Routines for Windows NT MD4 Hash functions
static int _my_wcslen(int16 * str)
{
    int len = 0;
    while (*str++ != 0)
        len++;
    return len;
}

/*------------------------------------*/
// ! Convert a string into an NT UNICODE string
/** Note that regardless of processor type
  * this must be in intel (little-endian)
  * format.
*/
static int _my_mbstowcs(int16 * dst, uchar * src, int len)
{
    int i;
    int16 val;
    for (i = 0; i < len; i++) {
        val = *src;
        SSVAL(dst, 0, val);
        dst++;
        src++;
        if (val == 0)
            break;
    }
    return i;
}

/*-----------------------------------------*/
// ! Calcula el NT/MD4 hash de una clave
int calcularNThash(uchar * passwd, uchar * MD4hash)
{
    int i;
    unsigned int uiLength = strlen(passwd);
    int16 bufferUnicode[LONG_PASSWD];
    _my_mbstowcs(bufferUnicode, passwd, uiLength);
/* Ensure string is null terminated */
    bufferUnicode[uiLength] = 0;
/* Calculate length in bytes */
    uiLength = _my_wcslen(bufferUnicode) * sizeof(int16);
    MDbegin(&MDContext);
    for (i = 0; i + 64 <= (signed) uiLength; i += 64)
        MDupdate(&MDContext, bufferUnicode + (i / 2), 512);
    MDupdate(&MDContext, bufferUnicode + (i / 2), (uiLength - i) * 8);
    memcpy(MD4hash, &(MDContext.buffer), 16);
//      mensaje("MD4 (%s) = ",passwd); for(i=0;i<16;i++)
//      mensaje("%02X",MD4hash[i]); mensaje("\n");
//      usleep(500000); 
    return OK;
}

/*-----------------------------------*/
// ! LEER ARGUMENTOS DE LA LINEA DE COMANDOS 
int leer_opciones(const int argc, char **argv)
{
    struct option opciones_largas[] = {
        {"help", no_argument, NULL, 'h'},
        {"proc", required_argument, NULL, 'n'},
        {"alfa", required_argument, NULL, 'a'},
        {"port", required_argument, NULL, 'p'},
        {"version", no_argument, NULL, 'v'},
        {"dict", no_argument, NULL, 'd'},
        {"log", required_argument, NULL, 'l'},
        {"verbose", no_argument, NULL, 'V'},
        {"frontend", required_argument, NULL, 'x'},
        {0, 0, 0, 0}            /* EL ULTIMO ELEMENTO DEBE LLENARSE DE CEROS */
    };
    int arg;
/* VALORES DE CONFIGURACION POR DEFECTO */
    parametros.procesos = CANT_PROCESOS_DEFAULT;
    parametros.verbose = !OK;
    parametros.arch_SAM = argv[argc - 1];
    parametros.arch_alfabeto = NULL;
    parametros.arch_log = NULL;
    parametros.arch_dict = NULL;
    parametros.puerto = PUERTO;
    parametros.frontend_host = NULL;
    parametros.frontend_puerto = -1;
/* DESHABILITAR MENSAJES DE ERROR DE PARAMETROS POR STDERR */
    // opterr = 1;
    if (argc == 1) {
#ifdef PACKAGE
        ayuda(PACKAGE);
#else
        ayuda(argv[0]);
#endif
        exit(OK);
    }
    while ((arg =
            getopt_long(argc, argv, "-p:a:n:l:d:x:hvV",
                        opciones_largas, NULL)) != -1) {
        switch (arg) {
        case 'n':
            if (optarg)
                if (atoi(optarg) >= 0)
                    parametros.procesos = atoi(optarg);
            break;
        case 'a':
            if (optarg)
                parametros.arch_alfabeto = optarg;
            break;
        case 'd':
            if (optarg)
                parametros.arch_dict = optarg;
            break;
        case 'l':
            if (optarg)
                parametros.arch_log = optarg;
            break;
        case 'p':
            if (optarg)
                if (atoi(optarg) >= 0)
                    parametros.puerto = atoi(optarg);
            break;
        case 'x':
            if (optarg) {
                parametros.frontend_host = strsep(&optarg, ":");
                parametros.frontend_puerto = atoi(optarg);
            }
            break;
        case 'h':
            ayuda(argv[0]);
            exit(OK);
        case 'v':
            version();
            exit(OK);
        case 'V':
            parametros.verbose = OK;
            break;
        case '?':
            ayuda(argv[0]);
            exit(OK);
            break;
//      default:
            // mensaje("?\? getopt() devolvio el caracter de codigo 
            // 0%o ?\?\n", arg);
        }
    }
    mensaje("Procesos: %d\n", parametros.procesos);
    if (parametros.arch_alfabeto != NULL)
        mensaje("Archivo alfabeto: %s\n", parametros.arch_alfabeto);
    if (parametros.arch_log != NULL)
        mensaje("Archivo log: %s\n", parametros.arch_log);
    if (parametros.puerto != 0)
        mensaje("Puerto de escucha: %d\n", parametros.puerto);
    return OK;
}

/*--------------------------------------*/

void ayuda(const char *nombre_propio)
{
    printf(_("CISILIA is a multiprocessing password cracking system.\n"));
    printf(_("Usage:\t%s [options] <password file>\n"), nombre_propio);
    printf(_("Options:\n"));
    printf(_("   -a, --alfa=file\t\tLoad custom alphabet from file "));
#if ALFABETO_NUMERICO == 0
    printf(_("(Default [A-Z])\n"));
#else
    printf(_("(Default [A-Z0-9])\n"));
#endif
    printf(_("   -d, --dict=file\t\tLoad dictionary file\n"));
    printf(_("   -h, --help\t\t\tShow this help message and exit\n"));
    printf(_("   -l, --log=file\t\tAppend cracked passwords to file\n"));
    printf(_
           ("   -n, --proc=num\t\tSet number of brute-force subprocesses\n"));
    printf(_("   -p, --port=num\t\tSet listening port\n"));
    printf(_("   -v, --version\t\tShow version information and exit\n"));
    printf(_("   -V, --verbose\t\tShow debugging information\n"));
    printf(_("   -x, --frontend=host:port\tConnect to a front-end\n"));
    printf("\n");
}

/*--------------------------------------*/

void version(void)
{
#ifdef VERSION
    printf("Version: %s\n", VERSION);
#else
    printf("Version: %s\n", "?.?");
#endif
}

/*--------------------------------------*/
/*
void imprimir_estadisticas(void)
{
    struct timeval t;
    double s;
    unsigned long h, m;
*/
        /*
         * struct rusage estado; HORAS(h), MINUTOS(m), SEGUNDOS(s) de
         * USUARIO(u) Y DE SISTEMA(s) double us, ss; unsigned long uh, um, sh, 
         * sm, h; struct tms tiempo;
         * 
         * mensaje("\n");
         * 
         * times(&tiempo); mensaje("utime= %ld\n", tiempo.tms_utime);
         * mensaje("cutime= %ld\n", tiempo.tms_cutime);
         * 
         * getrusage(RUSAGE_SELF, &estado); us = estado.ru_utime.tv_sec +
         * estado.ru_utime.tv_usec * .000001; um = us / 60; uh = um / 60; ss = 
         * estado.ru_stime.tv_sec + estado.ru_stime.tv_usec * .000001; sm = ss 
         * / 60; sh = sm / 60; mensaje("Tiempo de usuario:
         * %lu:%lu:%.03f\n", uh, um - uh * 60, us - um * 60);
         * mensaje("Tiempo de sistema: %lu:%lu:%.03f\n", sh, sm - sh *
         * 60, ss - sm * 60); mensaje("RSS maximo: %ld\n",
         * estado.ru_maxrss); 
         */
/*
    gettimeofday(&t, NULL);
    s = (t.tv_sec - datosPadre.ti.tv_sec) + (t.tv_usec - datosPadre.ti.tv_usec) * .000001;
    m = s / 60;
    h = m / 60;
    mensaje("\n");
    mensaje("Subproc: \t\t\t\t %d \n", pid);
    mensaje("Tiempo transcurrido: \t\t\t %luh%02lum%.03fs \n",
	    h, m - h * 60, s - m * 60);
    mensaje("Palabras validas generadas: \t\t %llu \n", datosPadre.palabras);
//    mensaje("Palabras invalidas generadas: \t\t %llu \n", invalidas);
//    mensaje("TOTAL Palabras generadas: \t\t %llu \n", palabras + invalidas);
    mensaje("TOTAL semiclaves descifradas: \t\t %llu \n", datosPadre.semicracks);
    mensaje("TOTAL CLAVES DESCIFRADAS: \t\t %llu \n", datosPadre.cracks);
    mensaje("Palabras validas generadas x segundo: \t %.3f \n",
	    datosHijo.palabras / s);
//    mensaje("Palabras generadas x segundo (crudo): \t %.3f \n",(palabras + invalidas) / s);
    mensaje("SemiClaves descifradas x segundo: \t %.3f \n",
	    datosPadre.semicracks / s);
    mensaje("Claves descifradas x segundo: \t\t %.3f \n", datosPadre.cracks / s);
    mensaje("RX Bytes: \t\t\t\t %lu \n", datosPadre.rx);
    mensaje("RX Bytes/seg: \t\t\t\t %.3f \n", datosPadre.rx / s);
}
*/
/*--------------------------------------*/

void capturar_senial(int senial)
{
    switch (senial) {
    case SIGSEGV:
        fprintf(stderr, "(PID:%d) ## SEGMENTATION FAULT ##\n", pid);
        exit(!OK);
        break;
    case SIGPIPE:
        error(ERROR_SOCKET_SIGPIPE);
        break;
    case SIGCHLD:
/* DEBEMOS IGNORAR A LOS HIJOS DETENIDOS POR SIGSTOP O SIGTSTP !!! */
        waitpid(-1, NULL, WNOHANG);
        break;
    default:
        mensaje("Senial no reconocida");
    }
}

/*--------------------------------------*/

void enviar_estadisticas()
{
    int i;
    int tx;

    if (sock_frontend > 0) {
        datosPadre.palabras = 0;
/* EMPEZAR EN i=0 PARA CONTAR LA PALABRA GENERADAS POR EL PROC NULO */
/* POR AHORA NO LAS CONTAMOS PORQUE EL PRIMER HIJO TAMBIEN LA GENERA */
        for (i = 0; i <= parametros.procesos; i++) {
            datosPadre.palabras += vhijos[i].palabras;
        }
        gettimeofday(&datosPadre.t, NULL);

        tx = send(sock_frontend, &datosPadre, sizeof(datosPadre), 0);
        if (tx < sizeof(datosPadre))
            advertencia("ERROR en transmision de estadisticas");
    }
}

/*--------------------------------------*/

void enviar_datos_hijo()
{
    int tx;
    tx = send(sockudp, &datosHijo, sizeof(datosHijo), 0);
    if (tx < sizeof(datosHijo))
        advertencia("ERROR enviando datos del hijo");
}

/*--------------------------------------*/

void inicioPrincipal(void)
{
    int i;

/* CAMBIAR A UN PROCESS GROUP ID NUEVO */
    setpgid(0, 0);

/* ESTABLECER LA JERARQUIA DEL PROCESO */
    jerarquia = PADRE;

/* ABRIR SOCKETS DE ESCUCHA TCP */
    socktcp = crearSocketEscuchaTCP(parametros.puerto);
    sockudp = crearSocketEscuchaUDP(parametros.puerto);

/* CAPTURAR SENIALES */
    signal(SIGTSTP, SIG_DFL);	/* NO HEREDAR EL MANEJADOR DEL FRONTEND */
    signal(SIGCONT, SIG_DFL);	/* NO HEREDAR EL MANEJADOR DEL FRONTEND */
    signal(SIGTTOU, SIG_IGN);   /* ESCRITURA EN BACKGROUND */
    signal(SIGSEGV, capturar_senial);
    signal(SIGPIPE, capturar_senial);
    signal(SIGCHLD, capturar_senial);
    signal(SIGUSR1, salir);
    signal(SIGQUIT, salir);
    signal(SIGTERM, salir);
    signal(SIGINT, salir);

/* INICIAR DATOS DEL PADRE */
    gettimeofday(&(datosPadre.ti), NULL);
    datosPadre.total_palabras = 0;
    datosPadre.lineas = 0;
    datosPadre.palabras = 0;
    datosPadre.semicracks = 0;
    datosPadre.cracks = 0;
    datosPadre.rx = 0;
//    datosPadre.invalidas = 0;

/* INICIAR DATOS DE LOS HIJOS */
    datosHijo.id_hijo = 0;
    datosHijo.palabras = 0;

/* RESERVAR ESTRUCTURAS DE DATOS */
#if CONEXION_PERMANENTE != 0
/* CONTAR EL PROCESO DEL DICCIONARIO */
    vclientes = (int *) malloc(sizeof(int) * (parametros.procesos + 1));
    if (vclientes == NULL)
        error(ERROR_MALLOC);
#endif
    vhijos = (datos_hijo *) malloc(sizeof(datos_hijo) *
                                   (parametros.procesos + 1));
    if (vhijos == NULL)
        error(ERROR_MALLOC);
    for (i = 0; i <= parametros.procesos; i++)
        vhijos[i].palabras = 0;

/* LEER CLAVES */
    datosPadre.lineas = archivo_a_memoria(parametros.arch_SAM);

/* GENERAR ESTRUCTURAS DEL ALFABETO */
    cargar_alfabeto(parametros.arch_alfabeto);
//    imprimir_datos_users_plano(); 
    ordenar_LMhashes(2 * datosPadre.lineas);
//    imprimir_datos_users_plano();
    generar_indice_LMhashes(2 * datosPadre.lineas);
    datosPadre.total_palabras = total_palabras_validas();
    mensaje("TOTAL PALABRAS VALIDAS = %llu\n\n",
            datosPadre.total_palabras);

/* CONECTARSE AL FRONTEND SI ES NECESARIO */
    if (parametros.frontend_host != NULL) {
        sock_frontend =
            conectarUDP(parametros.frontend_host,
                        parametros.frontend_puerto);
        if (sock_frontend > 0) {
            enviar_estadisticas();
            mensaje("FRONT-END EN %s:%d/UDP\n",
                    parametros.frontend_host, parametros.frontend_puerto);
            signal(SIGALRM, enviar_estadisticas);
            ualarm(TIEMPO_ESTADISTICAS, TIEMPO_ESTADISTICAS);
        } else
            advertencia("ERROR conectando al front-end");
    }
}

/*--------------------------------------*/

void inicioPadre(void)
{
    int i;
    
    signal(SIGUSR1, SIG_IGN);

    free(LMhashes);
    free(indice_LMhashes);
    for (i = 0; i < LONG_PASSWD / 2; i++)
        free(alfabeto[i]);
    for (i = 0; i < LONG_PASSWD / 2; i++)
        free(matriz_LMhashes[i]);
    free(matriz_LMhashes);
}

/*--------------------------------------*/

void inicioSubproc(void)
{
/* MANEJAR CIERTAS SENIALES */
    signal(SIGALRM, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);

/* ESTABLECER LA JERARQUIA DEL PROCESO */
    jerarquia = HIJO;

/* CERRAR LOS SOCKETS HEREDADOS DEL PADRE */
    cerrar_sockets();

/* LIBERAR DATOS DEL PADRE */
#if CONEXION_PERMANENTE != 0
    free(vclientes);
#endif
    free(vhijos);
    free(datosUsers);

/* INICIAR DATOS DEL PROCESO HIJO */
    pid = getpid();
    datosHijo.pid = pid;
    clientes = 0;

/* CONECTARSE AL PADRE */
#if CONEXION_PERMANENTE != 0
    socktcp = conectarTCP(LOCALHOST, parametros.puerto);
#endif
    sockudp = conectarUDP(LOCALHOST, parametros.puerto);

/* INICIAR ALARMA PARA ENVIAR ESTADISTICAS */
    ualarm(TIEMPO_ESTADISTICAS * parametros.procesos,
           TIEMPO_ESTADISTICAS * parametros.procesos);
    signal(SIGALRM, enviar_datos_hijo);
}

/*--------------------------------------*/

void matarSubProcesos(void)
{
    kill(0, SIGUSR1);
}

/*--------------------------------------*/

void cerrar_sockets(void)
{
#if CONEXION_PERMANENTE != 0
    int i;
    for (i = 0; i < clientes; i++)
        close(vclientes[i]);
#else
    close(cliente);
#endif
    close(socktcp);
    close(sockudp);
    close(sock_frontend);
}

/*--------------------------------------*/

void salir()
{
//    imprimir_estadisticas();
    mensaje("(PID:%d) SALIENDO...\n", pid);

    if (jerarquia == PADRE) {
/* AVISAR AL FRONTEND QUE YA SALIMOS */
        if (sock_frontend > 0)
            send(sock_frontend, NULL, 0, 0);
        matarSubProcesos();
        cerrar_sockets();
        exit(OK);
    } 

    if (jerarquia == HIJO) {
        cerrar_sockets();
        _exit(OK);
    }
}

/*--------------------------------------*/

int crearSocketEscuchaTCP(int puerto)
{
    int sockserver, ok = 1;
    struct sockaddr_in server;
    socklen_t longitud;
    sockserver = socket(PF_INET, SOCK_STREAM, 0);
    if (sockserver < 0)
        error(ERROR_SOCKET);
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);
    server.sin_port = htons(puerto);
    setsockopt(sockserver, SOL_SOCKET, SO_REUSEADDR, &ok, sizeof(ok));
    if (bind(sockserver, (struct sockaddr *) &server, sizeof server) < 0)
        error(ERROR_SOCKET_BIND);
    if (puerto == 0) {          /* SI EL PUERTO DE ESCUCHA ES ALEATORIO */
        getsockname(sockserver, (struct sockaddr *) &server, &longitud);
        parametros.puerto = ntohs(server.sin_port);
/* MOSTRAR EL PUERTO ALEATORIO ASIGNADO  */
        mensaje(_("Random port: "));
        mensaje("%d/tcp \n", parametros.puerto);
    }

    if (listen(sockserver, LISTEN_QUEUE) < 0)
        error(ERROR_SOCKET_LISTEN);
    return (sockserver);
}

/*--------------------------------------*/

int crearSocketEscuchaUDP(int puerto)
{
    int sockserver;
    int ok = 1;
    struct sockaddr_in server;
    socklen_t longitud;
    sockserver = socket(PF_INET, SOCK_DGRAM, 0);
    if (sockserver < 0)
        error(ERROR_SOCKET);
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = htonl(INADDR_ANY);
    server.sin_port = htons(puerto);
    setsockopt(sockserver, SOL_SOCKET, SO_REUSEADDR, &ok, sizeof(ok));
    if (bind(sockserver, (struct sockaddr *) &server, sizeof server) < 0)
        error(ERROR_SOCKET_BIND);
    if (puerto == 0) {          /* SI EL PUERTO DE ESCUCHA ES ALEATORIO */
        getsockname(sockserver, (struct sockaddr *) &server, &longitud);
        parametros.puerto = ntohs(server.sin_port);
/* MOSTRAR EL PUERTO ALEATORIO ASIGNADO  */
        mensaje(_("Random port: "));
        mensaje("%d/udp \n", parametros.puerto);
    }

    return (sockserver);
}

/*--------------------------------------*/

int conectarTCP(const char *ip_server, const int port_server)
{
    int sock;
    struct sockaddr_in server;
    struct hostent *hp;

    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0)
        error(ERROR_SOCKET);
    server.sin_family = AF_INET;
    hp = gethostbyname(ip_server);
    if (hp == NULL)
        error(ERROR_SOCKET_HOSTNAME);
    memcpy((char *) &server.sin_addr, (char *) hp->h_addr, hp->h_length);
    server.sin_port = htons(port_server);
    if (connect(sock, (struct sockaddr *) &server, sizeof server) < 0)
        error(ERROR_SOCKET_CONN);
    return (sock);
}

/*--------------------------------------*/

int conectarUDP(const char *ip_server, const int port_server)
{
    int sock;
    struct sockaddr_in server;
    struct hostent *hp;

    sock = socket(PF_INET, SOCK_DGRAM, 0);
    if (sock < 0)
        error(ERROR_SOCKET);
    server.sin_family = AF_INET;
    hp = gethostbyname(ip_server);
    if (hp == NULL) {
        perror("ERROR en el IP UDP");
        return -1;
    }
    memcpy((char *) &server.sin_addr, (char *) hp->h_addr, hp->h_length);
    server.sin_port = htons(port_server);
    if (connect(sock, (struct sockaddr *) &server, sizeof server) < 0) {
        perror("ERROR conectando al socket UDP");
        return -1;
    }
    return (sock);
}

/*--------------------------------------*/

static int escuchar()
{
    fd_set bolsa_lectura;
    struct timeval tv;
    int r = 0, rx_total = 0;
    int s, i;
    crack descifrado;
    msgHijo buffer;
    struct sockaddr_in dir;
    socklen_t longitud = sizeof(dir);

    FD_ZERO(&bolsa_lectura);
    FD_SET(socktcp, &bolsa_lectura);
    FD_SET(sockudp, &bolsa_lectura);
#if CONEXION_PERMANENTE != 0
    for (i = 0; i < clientes; i++)
        FD_SET(vclientes[i], &bolsa_lectura);
#endif
    tv.tv_sec = BUSY_WAIT_SEC;
    tv.tv_usec = BUSY_WAIT_USEC;

    do {
        errno = 0;
        s = select(FD_SETSIZE, &bolsa_lectura, NULL, NULL, &tv);
    } while ((s < 0) && (errno == EINTR));

    if (s < 0 && errno != EINTR)
        error(ERROR_SOCKET_SELECT);
    if (s == 0)
        return OK;
#if CONEXION_PERMANENTE != 0
    for (i = 0; i < clientes; i++)
        if (FD_ISSET(vclientes[i], &bolsa_lectura)) {

            do {
                errno = 0;
                r = recv(vclientes[i],
                         &(((char *) (&descifrado))[rx_total]),
                         sizeof(descifrado) - rx_total, MSG_WAITALL);
                rx_total += r;
            } while (rx_total < sizeof(descifrado) && r > 0);

            if (rx_total < sizeof(descifrado)) {
                mensaje("rx=%d de %d ", rx_total, sizeof(descifrado));
                if (rx_total < 0)
                    advertencia("ERROR en recv(): < 0");
                if (rx_total > 0)
                    advertencia("ERROR en recv(): < esperado");
//              if (rx_total == 0)
//                  mensaje("(PID:%d) ERROR en recv... (==0)\n",datosPadre.pid);
                close(vclientes[i]);
                clientes--;
                vclientes[i] = vclientes[clientes];
                return !OK;
            } else {
                datosPadre.rx += rx_total;
                if (recibirDatosUsers
                    (descifrado.semipasswd, descifrado.usuario,
                     descifrado.mitad) == OK) {
                    chequear_passwd_ok(descifrado.usuario);
                }
            }
        }

    if (FD_ISSET(socktcp, &bolsa_lectura)) {
        vclientes[clientes] = accept(socktcp, NULL, NULL);
        if (vclientes[clientes] < 0)
            error(ERROR_SOCKET_ACCEPT);
        clientes++;
        mensaje("Cliente aceptado\n");
    }
#else
    for (i = 0; i < s; i++) {
        cliente = accept(fd, NULL, NULL);
        if (cliente < 0)
            error(ERROR_SOCKET_ACCEPT);
//      do {
        errno = 0;
        r = recv(cliente, &descifrado, sizeof(descifrado), 0);
//      } while ((r < sizeof(descifrado)) && (errno == EINTR));
        if (r < 0)
            error(ERROR_SOCKET_RECV);
        if (r < sizeof(descifrado)) {
            advertencia("Cerrando conexion rota");
        } else {
            datosPadre.rx += r;
            if (recibirDatosUsers
                (descifrado.semipasswd, descifrado.usuario,
                 descifrado.mitad) == OK)
                chequear_passwd_ok(descifrado.usuario);
        }
        close(cliente);
    }
#endif

    if (FD_ISSET(sockudp, &bolsa_lectura)) {
        r = recvfrom(sockudp, &buffer, sizeof(buffer), 0,
                     (struct sockaddr *) &dir, &longitud);
        switch (r) {
        case sizeof(datos_hijo):
            vhijos[buffer.data.id_hijo] = buffer.data;
            break;
        default:
            mensaje("ERROR en mensaje de estadisticas del hijo.\n");
            break;
        }
    }

    return OK;
}

/*--------------------------------------*/

int crackear_passwords_dic(uchar * archivo)
{
    FILE *pdic;
/* RESERVAR ESPACIO PARA LA PALABRA, EL '\n' Y EL '\0' */
    char linea_dic[LONG_PASSWD + 2] = { '\0' };
    uchar semiLMhash[8] /*= { '\0' }*/ ;
    char palabra_nula[LONG_PASSWD / 2] = { '\0' };
    char *c;
    int i;
    int subpid;

    if (archivo != NULL) {
        switch (subpid = fork()) {

        case -1:               /* ERROR fork() */
            error(ERROR_FORK);

        case 0:                /* SUBPROCESO */
            inicioSubproc();
            signal(SIGALRM, SIG_IGN);
            ualarm(0, 0);

            if (parametros.procesos == 0) {
                mensaje("(PID:%d) GENERANDO PALABRA NULA...\n",
                        datosHijo.pid);
                E1(palabra_nula, CONSTANTE_MAGICA, semiLMhash);
                BUSCAR(semiLMhash, palabra_nula);
                datosHijo.palabras++;
                enviar_datos_hijo();
            }

            mensaje("(PID:%d) PROCESANDO DICCIONARIO...\n", datosHijo.pid);
            if ((pdic = fopen(archivo, "rt")) == NULL)
                error(ERROR_ARCH_DIC);
            while (fgets(linea_dic, LONG_PASSWD + 2, pdic) != NULL) {
/* PASAR A MAYUSCULAS Y REEMPLAZAR TODOS LOS '\n' POR '\0' */
                for (i = 0; i < strlen(linea_dic); i++)
                    linea_dic[i] = toupper(linea_dic[i]);
                while ((c = strchr(linea_dic, '\n')) != NULL)
                    *c = '\0';
//                      mensaje("linea_dic: %s ", linea_dic);
//                      mensaje(" ==> &linea_dic[7]: %s\n", &linea_dic[7]);
                E1(linea_dic, CONSTANTE_MAGICA, semiLMhash);
                BUSCAR(semiLMhash, linea_dic);
#if SUBPROC_GEN_PAL_NULA != 0
                if (strlen(linea_dic) > (LONG_PASSWD / 2)) {
#endif
                    E1(&linea_dic[7], CONSTANTE_MAGICA, semiLMhash);
                    BUSCAR(semiLMhash, &linea_dic[7]);
#if SUBPROC_GEN_PAL_NULA != 0
                }
#endif
/* INDISPENSABLE PARA BORRAR LA LINEA ANTERIOR!!! */
                memset(linea_dic, '\0', LONG_PASSWD + 1);
            }
//            imprimir_estadisticas();
            mensaje("(PID:%d) DICCIONARIO COMPLETO...\n", datosHijo.pid);
            salir();

        default:               /* PROCESO PADRE */
            escuchar();
        }
    }
    return OK;
}

/*--------------------------------------*/
/*
int crackear_passwords_palabra_nula (void)
{
    char palabra_nula[LONG_PASSWD / 2] = { '\0' };
    int subpid;
    uchar semiLMhash[8];

    switch (subpid = fork()) {
    case -1:
	error(ERROR_FORK);
    case 0:
	inicioSubproc();
	signal(SIGALRM, SIG_IGN);
	mensaje("(PID:%d) GENERANDO PALABRA NULA...\n", datosHijo.pid);
	E1(palabra_nula, CONSTANTE_MAGICA, semiLMhash);
	BUSCAR(semiLMhash, palabra_nula);
	datosHijo.palabras++;
	enviar_datos_hijo();
//	imprimir_estadisticas();
	salir();

    default:
		escuchar();
    }
    return OK;
}
*/
/*--------------------------------------*/

int mensaje(const char *formato, ...)
{
    int bytes = 0;
    va_list args;

    if (parametros.verbose == OK) {
//      fprintf (stderr, "(PID:%d) ",pid);
        va_start(args, formato);
        vfprintf(stderr, formato, args);
        va_end(args);
        fflush(stderr);
    }
    return bytes;
}

/*--------------------------------------*/

int advertencia(const char *formato)
{
    int bytes = 0;
//    va_list args;

//    if (parametros.verbose == OK) {
//      bytes += fprintf (stderr, "(PID:%d) ",pid);
//      va_start(args,formato);
//      bytes += vfprintf (stderr, formato, args);
//        va_end(args);
    bytes += fprintf(stderr, "\a(PID:%d) ", pid);
    fflush(stderr);
    perror(formato);
//    }
    return bytes;
}

/*--------------------------------------*/

int logear(const char *formato, ...)
{
    FILE *arch;
    int bytes = 0;
    va_list args;
    va_start(args, formato);
    vfprintf(stdout, formato, args);
    fflush(stdout);
    if (parametros.arch_log != NULL) {
        arch = fopen(parametros.arch_log, "a");
        if (arch == NULL) {
            advertencia("ERROR en archivo log");
            parametros.arch_log = NULL;
            va_end(args);
            return 0;
        }
        bytes += vfprintf(arch, formato, args);
        fclose(arch);
    }
    va_end(args);
    return bytes;
}

/*--------------------------------------*/
